/**
  *  \file smallset.h
  */
#ifndef STREU_CPLUSLIB_SMALLSET_H_INCLUDED
#define STREU_CPLUSLIB_SMALLSET_H_INCLUDED

/** Small set. This encapsulates a type-safe set where members are
    enumerators with numeric values corresponding to the bits of an
    unsigned long. */
template<typename T>
class SmallSet {
    unsigned long rep;

    SmallSet(unsigned long l, bool)
        : rep(l) { }
 public:
    /** Construct empty set. */
    SmallSet()
        : rep(0) { }

    /** Construct unit set. */
    explicit
    SmallSet(T t)
        : rep(1UL << t) { }

    /** Add an element. */
    SmallSet&
    operator |= (T t)
        { rep |= (1UL << t); return *this; }

    /** Add a set. */
    SmallSet&
    operator |= (SmallSet other)
        { rep |= other.rep; return *this; }

    SmallSet
    operator | (T t) const
        { return SmallSet<T>(*this) |= t; }

    SmallSet
    operator | (SmallSet other) const
        { return SmallSet<T>(*this) |= other; }

    /** Add an element (alternate name). */
    SmallSet&
    operator += (T t)
        { rep |= (1UL << t); return *this; }

    /** Add a set (alternate name). */
    SmallSet&
    operator += (SmallSet other)
        { rep |= other.rep; return *this; }

    SmallSet
    operator + (T t) const
        { return SmallSet<T>(*this) |= t; }

    SmallSet
    operator + (SmallSet other) const
        { return SmallSet<T>(*this) |= other; }

    /** Remove an element. */
    SmallSet&
    operator -= (T t)
        { rep &= ~(1UL << t); return *this; }

    /** Remove a set. */
    SmallSet&
    operator -= (SmallSet other)
        { rep &= ~other.rep; return *this; }

    SmallSet
    operator - (T t) const
        { return SmallSet<T>(*this) -= t; }

    SmallSet
    operator - (SmallSet other) const
        { return SmallSet<T>(*this) -= other; }

    /** Conjunction. */
    SmallSet&
    operator &= (SmallSet other)
        { rep &= other.rep; return *this; }

    SmallSet
    operator & (SmallSet other) const
        { return SmallSet<T>(*this) &= other; }

    /** Toggle an element. */
    SmallSet&
    operator ^= (T t)
        { rep ^= (1UL << t); return *this; }

    /** Toggle elements from other set. */
    SmallSet&
    operator ^= (SmallSet other)
        { rep ^= other.rep; return *this; }

    SmallSet
    operator ^ (T t) const
        { return SmallSet<T>(*this) ^= t; }

    SmallSet
    operator ^ (SmallSet other) const
        { return SmallSet<T>(*this) ^= other; }

    /** Test whether element is in. */
    bool
    contains(T t) const
        { return rep & (1UL << t); }

    /** Test whether this set contains all elements from the other one. */
    bool
    contains(SmallSet other) const
        { return (other.rep & ~rep) == 0; }

    /** Test whether this set contains any of the elements from the
        other one. */
    bool
    containsAnyOf(SmallSet other) const
        { return rep & other.rep; }

    /** Test whether this set is empty. */
    bool
    empty() const
        { return rep == 0; }

    /** Check whether this is a set containing at most one element. */
    bool
    isUnitSet() const
        { return (rep & (rep-1)) == 0; }

    /** Static constructor: construct a set which contains all objects
        from T(0) up to T(t) */
    static SmallSet
    allUpTo(T t)
        {
            // this relies on the guaranteed overflow behaviour
            return SmallSet<T>((1UL << (t+1)) - 1, true);
        }

    bool
    operator==(SmallSet other) const
        {
            return rep == other.rep;
        }

    bool
    operator!=(SmallSet other) const
        {
            return rep != other.rep;
        }

    bool
    nonempty() const
        {
            return rep;
        }

    unsigned long toInteger() const { return rep; }
    static SmallSet fromInteger(unsigned long v) { return SmallSet<T>(v, true); }
};

#endif
